/* --------------------------------------------------------------------------
  Test Program for AXM-560


  Memory Map
  ----------
  00000000 - 00FFFFFF Internal Flash

  20000000 - 203FFFFF CS0 - External RAM 
  20400000 - 207FFFFF unused (mirrored External RAM)

  20800000 - 20FFFFFF 
    20800000 - 20FFFFEF CS2 - TTP Offboard 

  21000000 - 217FFFFF 
    21000000 - 213FFFFF FPGA local SRAM
    21400000 - 2147FFFF FPGA local Flash - lower 512KB
    21480000 - 214FFFFF FPGA local Flash - upper
    21680000 - 21681FFF FPGA internal mem
    2173000C - ..            upper flash registers

  21800000 - 21FFFFFF 
  22000000 - 227FFFFF 

  22800000 - 22FFFFFF 
  23000000 - 237FFFFF 
 
  
  40000000 -           Core 0 SRAM
  50000000 -           Core 1 SRAM
  
  60000000 - 63FFFFFF  DRAM

-------------------------------------------------------------------------- */

#include "mpc5675k-2.04.h"
#include "common.h"
#include "serial.h"

// --------------------------------------------------------------------------

#define EBI_BASE_REG 0xC3F84000

#define EXT_SRAM_START  0x20000000
#define EXT_SRAM_SIZE   0x400000
#define EXT_SRAM_WORDS EXT_SRAM_SIZE / 4
#define EXT_SRAM (*(unsigned long int*)EXT_SRAM_START)

#define FPGA_SRAM_START  0x21000000
#define FPGA_SRAM_SIZE   0x400000
#define FPGA_SRAM_WORDS FPGA_SRAM_SIZE / 4
#define FPGA_SRAM (*(unsigned long int*)FPGA_SRAM_START)

#define FPGA_IRAM_START  0x21680000
#define FPGA_IRAM_SIZE   0x2000
#define FPGA_IRAM_WORDS FPGA_IRAM_SIZE / 4
#define FPGA_IRAM (*(unsigned long int*)FPGA_IRAM_START)

#define FPGA_FLASH_START  0x21400000
#define FPGA_FLASH_SIZE   0x100000
#define FPGA_FLASH_WORDS FPGA_FLASH_SIZE / 4
#define FPGA_FLASH (*(unsigned long int*)FPGA_FLASH_START)
//#define FPGA_FLASH_PAGE   0x2173000C
#define FPGA_FLASH_PAGE   0x2170000C

// 64-meg byte LPDDR
#define EXT_DRAM_START  0x60000000
#define EXT_DRAM_SIZE   0x4000000
//#define EXT_DRAM_SIZE   0x0000100		// test
#define EXT_DRAM_WORDS EXT_SRAM_SIZE / 4
#define EXT_DRAM (*(unsigned long int*)EXT_DRAM_START)
#define EXT_DRAM16 (*(unsigned short*)EXT_DRAM_START)


#define EBI_BUSMODE	0
#define DDR_BUSMODE	1

#define DRUN_MODE	3
#define VLE				2			 
#define DPM_KEY1	0x5AF0
#define DPM_KEY2	0xA50F


// --------------------------------------------------------------------------

extern void _start (void);	// first core startup
extern void __start_p1();		// second core startup 

void main0 (void);
void main1 (void);

void DISABLE_WATCHDOG(void);
void MC_MODE_INIT(void);
void SET_PLL_0(void);
void SET_PLL_1(void);
void CLKOUT(void);
void Toggle_LED_473(void);
void Toggle_LED_257(void);

static void PerformanceTest(void);


void setup_fmpll(void);
void initialize_hardware();
void SetBusMode(char mode);


void delay(int count);
uint8_t GetIdPort();

void FlexrayOpTest();
void LEDTest();
void ExpansionConTest();
void CANTest();
void PushSwitchTest();
void LANTest();

char LEDDelay();
void LEDDelay2();
void LEDDelay3();
char TogDelay();

void setup_EBI();
void ExternalRAMTest();
void EBIWriteTest();
void FPGATest();
void DDRTest();
void ETPUTest();
void ZigBeeTest();  
void LVDSTest();  



// --------------------------------------------------------------------------
uint8_t   CORE_ID;
uint8_t   error;

// --------------------------------------------------------------------------
//
// main() - called from __start in __start.c
//
// main C entry point for core 0

void main (void){

  volatile int i = 0;
//  char ch;


	DISABLE_WATCHDOG();	/* Disable SWT */
	
	RGM.FES.B.F_FCCU_SAFE = 0x1;  // Clear any fails Debug ONLY


	setup_fmpll();	// Setup PLL 
	CLKOUT();		// Output Selected Clock Source 

	// test - endless loop here
	while(1){

		// LEDs
	  SIU.PCR[48].R = 0x0200;		// Enable pins as outputs
	  SIU.PCR[49].R = 0x0200;	
	  SIU.PCR[60].R = 0x0200;		
	  SIU.PCR[98].R = 0x0200;	

		// Switches	
		SIU.PCR[78].R = 0x0103;	// Enable pins as inputs, push pull
		SIU.PCR[93].R = 0x0103;	

		// RS485
	  SIU.PCR[142].R = 0x0200;		// Enable pins as outputs
	  SIU.PCR[144].R = 0x0200;		
	  SIU.PCR[18].R = 0x0200;		
    SIU.GPDO[142].B.PDO = 1;  // make pin high
    SIU.GPDO[144].B.PDO = 1;  // make pin high

		// CAN
	  SIU.PCR[16].R = 0x0200;		// Enable pins as outputs


  	delay(10);  // must wait before reading

	  // continuously toggle pins and look for switch inputs
	  for( ; ;){

			// if SW1 down
		  if(SIU.GPDI[78].B.PDIN == 1)	SIU.GPDO[60].B.PDO = 1;  // make pin high
		  else													SIU.GPDO[60].B.PDO = 0;  // make pin low
			// if SW2 down
		  if(SIU.GPDI[93].B.PDIN == 1)	SIU.GPDO[98].B.PDO = 1;  // make pin high
		  else													SIU.GPDO[98].B.PDO = 0;  // make pin low

			// LEDs
	    SIU.GPDO[48].B.PDO = 1;  // make pin high
	    SIU.GPDO[49].B.PDO = 0;  // make pin low

			// RS485
	    SIU.GPDO[18].B.PDO = 1;  // make pin high

			// CAN
  	  SIU.GPDO[16].B.PDO = 1;  // make pin high


	    //---------
	    LEDDelay3(); 



			// if SW1 down
		  if(SIU.GPDI[78].B.PDIN == 1)	SIU.GPDO[60].B.PDO = 1;  // make pin high
		  else													SIU.GPDO[60].B.PDO = 0;  // make pin low
			// if SW2 down
		  if(SIU.GPDI[93].B.PDIN == 1)	SIU.GPDO[98].B.PDO = 1;  // make pin high
		  else													SIU.GPDO[98].B.PDO = 0;  // make pin low

			// LEDs
	    SIU.GPDO[49].B.PDO = 1;  // make pin high
	    SIU.GPDO[48].B.PDO = 0;  // make pin low

			// RS485
	    SIU.GPDO[18].B.PDO = 0;  // make pin low

			// CAN
  	  SIU.GPDO[16].B.PDO = 0;  // make pin low


	    //---------
	    LEDDelay3(); 
	  }
	}



//------LED test
  SIU.PCR[141].R = 0x0200;		// Enable pins as outputs
  // user SW_1 LED 1 should be on since it's active low by default

//while(1){;}	// endless loop
//------LED test

   
		
/*
	PerformanceTest();

	SSCM.DPMBOOT.R = (unsigned int) __start_p1 + VLE;       // start address for Core1
   	SSCM.DPMKEY.R  = DPM_KEY1;                          // kick starts core1             
	SSCM.DPMKEY.R  = DPM_KEY2;                          // with these values  
	
	Toggle_LED_473();


	while(1){;} // Wait Forever 
*/
	

	// gpio[92] should be low to boot from flash

/*
// test - endless loop here
while(1){	 ; }

//------LED test
  SIU.PCR[141].R = 0x0200;		// Enable pins as outputs
  SIU.PCR[142].R = 0x0200;	
  // continuously toggle pins until user presses a key
  for( ; ;){
    SIU.GPDO[141].B.PDO = 1;  // make pin high
    SIU.GPDO[142].B.PDO = 0;  // make pin low
    LEDDelay2(); 
    SIU.GPDO[142].B.PDO = 1;  // make pin high
    SIU.GPDO[141].B.PDO = 0;  // make pin low
    LEDDelay2(); 
  }
//------LED test

*/

/*   
	initialize_hardware();

  GetIdPort();  // read the controller number, store in CORE_ID

  // Start the second core, VLE mode
  printf("\nStarting 2nd Core ..");
//  SSCM.DPMBOOT.R = (unsigned long)__start_p1 + 0x00000002; 
//  SSCM.DPMKEY.R = 0x00005AF0;
//  SSCM.DPMKEY.R = 0x0000A50F;
	SSCM.DPMBOOT.R = (unsigned int) __start_p1 + VLE;   // start address for Core1
 	SSCM.DPMKEY.R  = DPM_KEY1;                          // kick starts core1             
	SSCM.DPMKEY.R  = DPM_KEY2;                          // with these values  


  // Loop forever 

  // main loop
	for (;;) {
		printf("\n\nAXM-536 (PMC-5675) Diagnostic Program  --  %s\n\n", __DATE__);

		printf("  1 Expansion Connection\n");
		printf("  2 Flexray\n");
		printf("  3 CAN\n");
		printf("  4 Push Switch\n");
		printf("  5 LEDs\n");
		printf("  6 LAN\n");
		printf("  7 External Memory\n");
//		printf("  7 EBI - Write Test\n");

		printf("  8 FPGA Memory\n");
		printf("  9 DDR Memory\n");
//		printf("  9 DSPI (all controllers)\n");
//		printf("  L LVDS\n");
//		printf("  Z ZigBee\n");
//		printf("  E EBI Mode\n");
//		printf("  D DDR Mode\n");

		printf("\n  0 Test All\n");

    printf("\n[C%u] >", (CORE_ID + 1)); // prompt
    
    ch = getchar();	// wait for a character from serial port
    putchar(ch);        // echo it
    
    switch (ch){
      case '1': ExpansionConTest(); break;
      case '2': FlexrayOpTest();  break;
      case '3': CANTest();  break;
      case '4': PushSwitchTest(); break;
      case '5': LEDTest();  break;
      case '6': LANTest();  break;
      case '7': ExternalRAMTest();  break;
//      case '7': EBIWriteTest();  break;
      case '8': FPGATest();  break;
      case '9': DDRTest();  break;
//      case '9': DSPITest();  break;
//      case 'l': case 'L': LVDSTest();  break;
//      case 'z': case 'Z': ZigBeeTest();  break;

      case 'e': case 'E': 
//				SetBusMode(EBI_BUSMODE);	// Set bus to EBI mode
				break;
      case 'd': case 'D': 
//				SetBusMode(DDR_BUSMODE);	// Set bus to DDR mode
				break;
				
      case '0': 
        ExpansionConTest();
        FlexrayOpTest();
        CANTest();
        PushSwitchTest();
        LEDTest();
				LANTest();
        ExternalRAMTest();
        FPGATest();
        DDRTest();
//				DSPITest();
//				LVDSTest();
//      	ZigBeeTest();  
        break;

    	default:	printf("Unknown Command");					break;
    }
    putchar('\n');
	}
*/
} 


/****************************************************************
* FUNCTION : Toggle_LED for 473BGA EVB
* DESCRIPTION : Configures and toggles the motherboard LEDs on Port D4,5,6&7
* INPUTS : None
* OUTPUTS : None
****************************************************************/
void Toggle_LED_473(void)
{
    int DELAY;

    /* Configure Ports  as outputs to toggle LEDS */
	SIU.PCR[59].R = 0x0220;  //FLEXPWM0_B0 Ball=P1
	SIU.PCR[62].R = 0x0220;  //FLEXPWM0_B1 Ball=P2
	SIU.PCR[99].R = 0x0220;  //FLEXPWM0_A2 Ball=P2
	SIU.PCR[102].R = 0x0220;  //FLEXPWM0_A3 Ball=P4
	
	/* Toggle LEDs) */
while(1){
	SIU.GPDO[59].B.PDO ^= 1;  //FLEXPWM0_B0 Ball=P1
	SIU.GPDO[62].B.PDO ^= 1;  //FLEXPWM0_B1 Ball=P2
	SIU.GPDO[99].B.PDO ^= 1;  //FLEXPWM0_A2 Ball=P3
	SIU.GPDO[102].B.PDO ^= 1;  //FLEXPWM0_A3 Ball=P4
	
	for(DELAY=0;DELAY<=0x2FFFF;DELAY++){}	/* Allows the LED to be seen on the motherboard */

	}
}
/****************************************************************
* FUNCTION : Toggle_LED for 257BGA EVB
* DESCRIPTION : Configures and toggles the motherboard LEDs on Port D4,5,6&7
* INPUTS : None
* OUTPUTS : None
****************************************************************/
void Toggle_LED_257(void)
{
    int DELAY;

    /* Configure Ports  as outputs to toggle LEDS */
	SIU.PCR[148].R = 0x0220;  //FLEXPWM0_B0 Ball=P1
	SIU.PCR[150].R = 0x0220;  //FLEXPWM0_B1 Ball=P2
	SIU.PCR[153].R = 0x0220;  //FLEXPWM0_A2 Ball=P2
	SIU.PCR[151].R = 0x0220;  //FLEXPWM0_A3 Ball=P4
	
	/* Toggle LEDs) */
while(1){
	SIU.GPDO[148].B.PDO ^= 1;  //FLEXPWM0_B0 Ball=P1
	SIU.GPDO[150].B.PDO ^= 1;  //FLEXPWM0_B1 Ball=P2
	SIU.GPDO[153].B.PDO ^= 1;  //FLEXPWM0_A2 Ball=P3
	SIU.GPDO[151].B.PDO ^= 1;  //FLEXPWM0_A3 Ball=P4
	
	for(DELAY=0;DELAY<=0x2FFFF;DELAY++){}	/* Allows the LED to be seen on the motherboard */

	}
}

static void PerformanceTest(void)
{
    volatile int i;
    
    while(1)
    {
      // Toggle the pin to evaluate the cycle the loop consumes.  
      SIU.GPDO[148].B.PDO ^= 1;  //FLEXPWM0_B0 Ball=P1
   	  for (i = 0; i<1000;i++){;}    
    }
}




static const char *TestNames[] = {
  "",
  "Flexray Optical Module",
  "LED's",
  "Expansion Connection",
  "CAN A & B",
  0
};
static const char *TestPBGA[] = {
  "",
  "SCI2: TXD  RXD\n   FRO_IO: 0-7\n   FR_A: BGE  EN  TX  TXEN\n   FR_B: BGE  EN  TX  TXEN\n   DSPI_A: SCK  SIN  SOUT  CS0",
  "LED_1 & LED_2",
  "SCI_B: TXD  RXD\n   DSPI_C: SCK  SIN  SOUT  CS0  CS1\n   FP0A0-2, FP1X1-3, FP1A1, FP1A3, FP1B1, FP1B3, FPX02, FPX03",
  "CAN_A: TX  RS\n   CAN_B: TX  RS",
  0
};

 
char TestPrompt(char p){
  puts("\n\n\n------ ");
  puts((char *)TestNames[p]);
  puts("\n\nVerify these are toggling:\n   ");
  puts((char *)TestPBGA[p]);
  puts("\n\npress any key to exit..");
}


// --------------------------------------------------------------------------
void GPIOTest(char p, uint8_t *PNum){
  uint8_t i;
  	
  // Enable pins as outputs
  for(i=0; *(PNum+i); i++){
	  SIU.PCR[*(PNum+i)].R = 0x0200;	
	}
   
  delay(10);  

  TestPrompt(p);

  // continuously toggle pins until user presses a key
  for( ; ;){
    for(i=0; *(PNum+i); i++)
      SIU.GPDO[*(PNum+i)].B.PDO = 1;   // make pin high
    if(TogDelay()) return; 

    for(i=0; *(PNum+i); i++)
      SIU.GPDO[*(PNum+i)].B.PDO = 0;   // make pin low
    if(TogDelay()) return; 
  }
}

// --------------------------------------------------------------------------
void CANTest(){
  uint8_t PNum[] = {16, 14, 139, 140, 0};
  GPIOTest(4, PNum);
}

// --------------------------------------------------------------------------
void FlexrayOpTest(){
  uint8_t PNum[] = {57, 60, 48, 47, 43, 134, 51, 52, 44, 137, 37, 39, 38, 36, 54, 53, 102, 59, 62, 100, 103, 116, 0};
  GPIOTest(1, PNum);

}

// --------------------------------------------------------------------------
void ExpansionConTest(){
  
  uint8_t PNum[] = {94, 95, 11, 13, 12, 10, 9, 58, 80, 99, 119, 122, 125, 120, 126, 121, 127, 98, 101, 0};
  GPIOTest(3, PNum);
}


// --------------------------------------------------------------------------
void LEDTest(){
  
  SIU.PCR[141].R = 0x0200;		// Enable pins as outputs
  SIU.PCR[142].R = 0x0200;	
   
  delay(10);  

  TestPrompt(2);

	// LED's are active LOW

  // continuously toggle pins until user presses a key
  for( ; ;){
    SIU.GPDO[141].B.PDO = 1;  // make pin high (off)
    SIU.GPDO[142].B.PDO = 0;  // make pin low	 (on)

    if(LEDDelay()) return; 

    SIU.GPDO[142].B.PDO = 1;  // make pin high (off)
    SIU.GPDO[141].B.PDO = 0;  // make pin low  (on)

    if(LEDDelay()) return; 
  }
}


// --------------------------------------------------------------------------
void PushSwitchTest(){
  
  puts("\n\n\n------ Push Switch");

	// Enable pins as inputs, push pull
	SIU.PCR[93].R = 0x0103;	
  delay(10);  // must wait before reading

  if(SIU.GPDI[93].B.PDIN == 0){

    puts("\n FAIL - button input is held down");
    return;
  }

  puts("\n\nPress Switch (Cx_SW1)");
  puts("\n\npress any key to exit..");
  while(SIU.GPDI[93].B.PDIN == 1){
    if(kbhit()){  // return if user aborts
      getchar();
      return;
    }
  }
  puts("\nPASSED - Switch Press detected");
}

// --------------------------------------------------------------------------
void LANTest(){
 
  printf("\nTesting Ethernet Connection...");
 	enettest(); 
  
  printf("\n   Verify LAN LED's are ON then press a key to continue...");
  (void) getchar();     // wait for a char from the keyboard
}


// --------------------------------------------------------------------------
// Set up the CORE_ID port and return the CORE_ID code [0..3]
// --------------------------------------------------------------------------
uint8_t GetIdPort(){

	SIU.PCR[226].R = 0x0103;	// Enable CORE_ID pins as inputs, push pull
	SIU.PCR[227].R = 0x0103;
  delay(20);  // must wait before reading

	CORE_ID = SIU.GPDI[227].B.PDIN;	// get first bit
  CORE_ID <<= 1;
  CORE_ID += SIU.GPDI[226].B.PDIN;	// get second bit

	return (CORE_ID);
}



// --------------------------------------------------------------------------
// simple while loop delay
void delay(int count){
  while(count){
    --count;
  }
}

// --------------------------------------------------------------------------
//  Delay used between pin High/Low toggles
//  returns 0 if NO key pressed (user interrupted)
//  otherwise returns the key pressed ASCII code
// --------------------------------------------------------------------------
char TogDelay(){

  int count=10000;
  while(count){
    if(kbhit()){
      return getchar();
    }
    --count;
  }
  return 0;
}
// --------------------------------------------------------------------------
//  Delay used for LED blinking
// --------------------------------------------------------------------------
char LEDDelay(){ 
  uint32_t count, count2;

  for(count=50; count; count--){
    for(count2=0x3000; count2; count2--){
      if(kbhit()){
        return getchar();
      }
    }
  }
  return 0;
}
// same as previous but without serial input test
void LEDDelay2(){ 
  uint32_t count, count2;

  for(count=50; count; count--){
    for(count2=0x3000; count2; count2--){;}
  }
}
// 10x longer
void LEDDelay3(){ 
  uint32_t count, count2;

  for(count=500; count; count--){
    for(count2=0x3000; count2; count2--){;}
  }
}

// --------------------------------------------------------------------------
void EBIWriteTest(){
  int x;
  unsigned long int *mem_ptr;


  printf("\n   Writing 55555555 from %08x - %08x", &EXT_SRAM, EXT_SRAM_START + (EXT_SRAM_SIZE-1)); 
  
  mem_ptr = &EXT_SRAM;
  for (x = 0; x < EXT_SRAM_WORDS; x++){
	  *mem_ptr = 0x55555555;
	  ++mem_ptr;
  }

}

// --------------------------------------------------------------------------
void ExternalRAMTest(){
  int x;
  unsigned long int *mem_ptr;
  unsigned long int wval, rval;
  uint8_t *mem_ptrB;
  uint8_t wvalb, rvalb;

//  unsigned int *mem_ptr;
//  unsigned int read_val;
/*
	asm {
    lis r8, 0x200A
    ori r8, r8,0xAAA4
    lwz r6, 0(r8)    

    lis r8, 0x2005
    ori r8, r8,0x5554
    lwz r6, 0(r8)    

    lis r8, 0x200A
    ori r8, r8,0xAAA8
    lwz r6, 0(r8)    

    lis r0, 0x1234
    ori r0, r0,0x5678
    stmw r0, 0(r8)

    lwz r6, 0(r8)    
  }
*/  
	SetBusMode(EBI_BUSMODE);	// Set bus to EBI mode

  error = 0;

  printf("\nTesting all 00000000.."); 
  printf("\n   Writing: %08x - %08x", &EXT_SRAM, EXT_SRAM_START + (EXT_SRAM_SIZE-1)); 
  
  // write then read entire array
  mem_ptr = &EXT_SRAM;
  for (x = 0; x < EXT_SRAM_WORDS; x++){
	  *mem_ptr = 0x00;
	  ++mem_ptr;
  }

  printf("\n   Reading: %08x - %08x", &EXT_SRAM, EXT_SRAM_START + (EXT_SRAM_SIZE-1)); 
  mem_ptr = &EXT_SRAM;
  for (x = 0; x < EXT_SRAM_WORDS; x++){
	  rval = *mem_ptr;
	  if (rval != 0x00){
	    ++error;
	    printf("\naddr: %08x    wrote %08x  read %08x", mem_ptr, 0, rval); 
	  }
  	++mem_ptr;
  	if(error > 8){
	    printf("\nAborting test.");
  	  break;  // abort if too many errors
  	}
  }

  printf("\nTesting all FFFFFFFF.."); 
  printf("\n   Writing: %08x - %08x", &EXT_SRAM, EXT_SRAM_START + (EXT_SRAM_SIZE-1)); 
  
  // write then read entire array
  mem_ptr = &EXT_SRAM;
  for (x = 0; x < EXT_SRAM_WORDS; x++){
	  *mem_ptr = 0xFFFFFFFF;
	  ++mem_ptr;
  }

  printf("\n   Reading: %08x - %08x", &EXT_SRAM, EXT_SRAM_START + (EXT_SRAM_SIZE-1)); 
  mem_ptr = &EXT_SRAM;
  for (x = 0; x < EXT_SRAM_WORDS; x++){
	  rval = *mem_ptr;
	  if (rval != 0xFFFFFFFF){
	    ++error;
	    printf("\naddr: %08x    wrote %08x  read %08x", mem_ptr, 0xFFFFFFFF, rval); 
	  }
  	++mem_ptr;
  	if(error > 8){
	    printf("\nAborting test.");
  	  break;  // abort if too many errors
  	}
  }


  printf("\nTesting Ones Complement of the address.."); 
  printf("\n   Writing: %08x - %08x", &EXT_SRAM, EXT_SRAM_START + (EXT_SRAM_SIZE-1)); 
  
  // write then read entire array
  mem_ptr = &EXT_SRAM;
  for (x = 0; x < EXT_SRAM_WORDS; x++){
		// write ones complement of the address
    wval = (unsigned long int) mem_ptr;
		wval = ~wval;	
	  *mem_ptr = wval;
	  ++mem_ptr;
  }

  printf("\n   Reading: %08x - %08x", &EXT_SRAM, EXT_SRAM_START + (EXT_SRAM_SIZE-1)); 
  mem_ptr = &EXT_SRAM;
  for (x = 0; x < EXT_SRAM_WORDS; x++){
	  rval = *mem_ptr;
    wval = (unsigned long int) mem_ptr;
		wval = ~wval;	
	  if (rval != wval){
	    ++error;
	    printf("\naddr: %08x    wrote %08x  read %08x", mem_ptr, wval, rval); 
	  }
  	++mem_ptr;
  	if(error > 8){
	    printf("\nAborting test.");
  	  break;  // abort if too many errors
  	}
  }
  
  // test as bytes  
  
  printf("\nTesting 8-bit r/w.."); 
  printf("\n   Writing: %08x - %08x", &EXT_SRAM, EXT_SRAM_START + (EXT_SRAM_SIZE-1)); 

  mem_ptrB = (unsigned char *) &EXT_SRAM;
  for(x = 0; x < EXT_SRAM_SIZE; x++){
		// write ones complement of the address
    wvalb = (unsigned char)(unsigned long int) mem_ptrB;
		wvalb = (uint8_t) ~wvalb;	
	  *mem_ptrB = wvalb;
	  ++mem_ptrB;
  }
  printf("\n   Reading: %08x - %08x", &EXT_SRAM, EXT_SRAM_START + (EXT_SRAM_SIZE-1)); 
  mem_ptrB = (unsigned char *) &EXT_SRAM;
  for (x = 0; x < EXT_SRAM_SIZE; x++){
	  rvalb = *mem_ptrB;
    wvalb = (unsigned char)(unsigned long int) mem_ptrB;
		wvalb = (uint8_t) ~wvalb;	
	  if (rvalb != wvalb){
	    ++error;
	    printf("\naddr: %08x    wrote %08x  read %08x", mem_ptrB, wvalb, rvalb); 
	  }
  	++mem_ptr;
  	if(error > 8){
	    printf("\nAborting test.");
  	  break;  // abort if too many errors
  	}
  }
  

  printf("\n\n--- External Memory Test:  %d  Errors\n", error);
}


/********************************************************************/

#define CFI_INTEL_CMD_CLEAR_STATUS 0x00500050
#define CFI_INTEL_CMD_LOCK_SETUP   0x00600060
#define CFI_INTEL_CMD_UNLOCK_BLOCK 0x00D000D0
#define	CFI_INTEL_CMD_READ_STATUS	 0x00700070
#define	CFI_INTEL_CMD_BLOCK_ERASE	 0x00200020
#define	CFI_INTEL_CMD_CONFIRM			 0x00D000D0	
#define	CFI_INTEL_CMD_PROGRAM1		 0x00400040
#define	CFI_INTEL_CMD_PROGRAM2		 0x00100010
#define	CFI_INTEL_CMD_READ_ARRAY	 0x00FF00FF

#define	CFI_INTEL_SR_READY			    0x00800080
#define	CFI_INTEL_SR_ERASE_SUSPEND	0x00400040
#define	CFI_INTEL_SR_ERASE_ERROR		0x00200020
#define	CFI_INTEL_SR_PROGRAM_ERROR	0x00100010
#define	CFI_INTEL_SR_VPEN_ERROR			0x00080008
#define	CFI_INTEL_SR_PROGRAM_SUSPEND 0x00040004
#define	CFI_INTEL_SR_BLOCK_LOCKED		0x00020002
#define	CFI_INTEL_SR_BEFP			      0x00010001


// return device status word
unsigned long int FPGA_flash_status(){

  unsigned long int *ptr;

  ptr = (unsigned long int *) FPGA_FLASH_START;
//  printf("\nstatus: %08x = %08x", ptr, *ptr); 
  return(*ptr);
}

// place device in read mode
void FPGA_flash_readarray32(){

  unsigned long int *ptr;
  ptr = (unsigned long int *) FPGA_FLASH_START;
  *ptr = CFI_INTEL_CMD_READ_ARRAY;
}

// set the current page number (0-7)
void FPGA_flash_setpage32(int x){
  char p;
  unsigned long int pv;
  

  unsigned long int *ptr;
  ptr = (unsigned long int *) FPGA_FLASH_PAGE;

  // determine page value
  for(p=0, pv=0; p<8; p++, pv+= 0x100000){
    if(p==x){
      //printf("   addr: %08x   val: %08x", ptr, pv); 
      *ptr = pv;  // write page value
      return;
    }
  }

}



// read status and wait for flash ready
int FPGA_flash_ready(){

  unsigned long int sr;

  // following is automatic after unlock, program, erase
  // *ptr = CFI_INTEL_CMD_READ_STATUS;  // enter status register mode

  // assume we're in status mode already

  // wait for flash ready
  for(; ;){ 
    sr = FPGA_flash_status();        // read status
    if( (sr & CFI_INTEL_SR_READY) == CFI_INTEL_SR_READY) break;

    // todo: return 1 on timeout
  }  

	return 0;
}

int FPGA_flash_unlock_block32(unsigned long int addr){

  unsigned long int *ptr;
  //printf("\nunlock.."); 

  ptr = (unsigned long int *) addr;
  
  *ptr = CFI_INTEL_CMD_CLEAR_STATUS;  // clear status register
  *ptr = CFI_INTEL_CMD_LOCK_SETUP;    // prepare to unlock
  *ptr = CFI_INTEL_CMD_UNLOCK_BLOCK;  // unlock block

  return FPGA_flash_ready(); // wait for command to finish
}

int FPGA_flash_erase_block32(unsigned long int addr){

  unsigned long int *ptr;
  //printf("\nerase.."); 

  ptr = (unsigned long int *) addr;

  *ptr = CFI_INTEL_CMD_CLEAR_STATUS;  // clear status register
  *ptr = CFI_INTEL_CMD_BLOCK_ERASE;   // erase block
  *ptr = CFI_INTEL_CMD_CONFIRM;

  return FPGA_flash_ready(); // wait for command to finish
}

int FPGA_flash_program32(unsigned long int addr, unsigned long int data){

  unsigned long int *ptr;
  unsigned long int sr;
  //printf("\nprogramming %08x at addr %08x..", data, addr); 

  ptr = (unsigned long int *) addr;

  *ptr = CFI_INTEL_CMD_CLEAR_STATUS;  // clear status register
  *ptr = CFI_INTEL_CMD_PROGRAM1;      // word program
  *ptr = data;  // data to program
  
//  FPGA_flash_ready(); // wait for command to finish

  // wait for flash ready
  for(; ;){ 
    sr = FPGA_flash_status();        // read status
//	  printf("\nstatus = %08x   ", sr); 

    if( (sr & CFI_INTEL_SR_READY) == CFI_INTEL_SR_READY) break;

    // todo: return 1 on timeout
  }  

  sr = FPGA_flash_status();        // read status
  printf("\nstatus after program = %08x   ", sr); 
  
  // check for programming errors

  return 0;
  
}


static const unsigned long int WritePattern1[] = {
  0x11111111,  0x22222222,  0x44444444,  0x88888888,
};

// --------------------------------------------------------------------------
void FPGATest(){

  int x, p;
  unsigned long int *mem_ptr;
  unsigned long int rval, pattern;


	SetBusMode(EBI_BUSMODE);	// Set bus to EBI mode

  error = 0;


  printf("\n\nFPGA - Internal RAM:");
  printf("\nTesting alternating 55555555|AAAAAAAA .."); 
//  MemTest_Single32(0x00, &FPGA_IRAM, FPGA_IRAM_START + (FPGA_IRAM_SIZE-1));
  
 
  printf("\n   Writing: %08x - %08x", &FPGA_IRAM, FPGA_IRAM_START + (FPGA_IRAM_SIZE-1)); 
  
  // write then read entire array
  mem_ptr = &FPGA_IRAM;
  for (x = 0; x < FPGA_IRAM_WORDS-1; x+=2){
	  *mem_ptr = 0x55555555;
	  ++mem_ptr;
	  *mem_ptr = 0xAAAAAAAA;
	  ++mem_ptr;
  }

  printf("\n   Reading: %08x - %08x", &FPGA_IRAM, FPGA_IRAM_START + (FPGA_IRAM_SIZE-1)); 
  mem_ptr = &FPGA_IRAM;
  for (x = 0; x < FPGA_IRAM_WORDS-1; x+=2){
	  rval = *mem_ptr;
	  if (rval != 0x55555555){
	    ++error;
	    printf("\naddr: %08x    wrote %08x  read %08x", mem_ptr, 0x55555555, rval); 
	  }
  	++mem_ptr;

	  rval = *mem_ptr;
	  if (rval != 0xAAAAAAAA){
	    ++error;
	    printf("\naddr: %08x    wrote %08x  read %08x", mem_ptr, 0xAAAAAAAA, rval); 
	  }
  	++mem_ptr;

  	if(error > 8){
	    printf("\nAborting test.");
  	  break;  // abort if too many errors
  	}
  }
  

  printf("\n\nFPGA - SRAM:");
  
  printf("\n   Writing: %08x - %08x", &FPGA_SRAM, FPGA_SRAM_START + (FPGA_SRAM_SIZE-1)); 
  // write then read entire array
  mem_ptr = &FPGA_SRAM;
  for (x = 0; x < FPGA_SRAM_WORDS-1; x+=2){
	  *mem_ptr = 0xAAAAAAAA;
	  ++mem_ptr;
	  *mem_ptr = WritePattern1[CORE_ID];
	  ++mem_ptr;
  }

  printf("\n   Reading: %08x - %08x", &FPGA_SRAM, FPGA_SRAM_START + (FPGA_SRAM_SIZE-1)); 
  mem_ptr = &FPGA_SRAM;
  for (x = 0; x < FPGA_SRAM_WORDS-1; x+=2){
	  rval = *mem_ptr;
	  if (rval != 0xAAAAAAAA){
	    ++error;
	    printf("\naddr: %08x    wrote %08x  read %08x", mem_ptr, 0xAAAAAAAA, rval); 
	  }
  	++mem_ptr;

	  rval = *mem_ptr;
	  if (rval != WritePattern1[CORE_ID]){
	    ++error;
	    printf("\naddr: %08x    wrote %08x  read %08x", mem_ptr, WritePattern1[CORE_ID], rval); 
	  }
  	++mem_ptr;

  	if(error > 8){
	    printf("\nAborting test.");
  	  break;  // abort if too many errors
  	}
  }


  printf("\n\nFPGA - Flash:");

  pattern = 0x12003400;
  
  for(p=0; p<8; p++){
    printf("\n\n---- Page %d", p);
    FPGA_flash_setpage32(p);


    mem_ptr = (unsigned long int *) FPGA_FLASH_START;
    //printf("\naddr: %08x = %08x", mem_ptr, *mem_ptr); 

    printf("\naddr: %08x    read %08x   Erasing..", mem_ptr, *mem_ptr); 

    FPGA_flash_unlock_block32(FPGA_FLASH_START);  // unlock block
    FPGA_flash_erase_block32(FPGA_FLASH_START);   // erase block
    FPGA_flash_readarray32(); // go back to read mode
    printf("\naddr: %08x    read %08x   Programming %08x ..", mem_ptr, *mem_ptr, pattern); 

    FPGA_flash_program32(FPGA_FLASH_START, pattern); // program word
    FPGA_flash_readarray32(); // go back to read mode

    rval = *mem_ptr;
    printf("\naddr: %08x    read %08x", mem_ptr, rval); 
    if(rval == pattern) printf("   ok");
    else printf("  ------ ERROR");
    
    pattern += 0x00020001;

  }

 
	if(error)  printf("\n\n--- FPGA Memory Test:  %d  Errors\n", error);
}

// --------------------------------------------------------------------------
//#define EXT_DRAM_START  0x60000000
//#define EXT_DRAM_SIZE   0x4000000
//#define EXT_DRAM_WORDS EXT_SRAM_SIZE / 4
//#define EXT_DRAM (*(unsigned long int*)EXT_DRAM_START)

void DDRTest(){

  int x;
//  char ch;
  
  
//  unsigned long int *mem_ptr;
//  unsigned short *mem_ptr16;
  vuint32_t *mem_ptr;
  vuint16_t	*mem_ptr16;
  unsigned long int rval;

  error = 0;
 
  
  printf("\n\nDDR - Internal DRAM:");

  printf("\nInitializing .."); 
	SetBusMode(DDR_BUSMODE);	// Set bus to DDR mode


/*

	printf("\nPress 0 to exit..");
	while(1){
		
	  mem_ptr = &EXT_DRAM;

	  printf("\n\naddr: %08x = %08x", mem_ptr, *mem_ptr++); 
	  printf("\naddr: %08x = %08x", mem_ptr, *mem_ptr); 

	  mem_ptr = &EXT_DRAM;

	  *mem_ptr++ = 0x55555555;
	  *mem_ptr = 0xAAAAAAAA;
	  
	  //printf("\npress a key to continue.."); 
    getchar();	// wait for a character from serial port
    ch = getchar();	// wait for a character from serial port
    if(ch == '0') return;
	}

*/

  printf("\nTesting alternating 55555555|AAAAAAAA .."); 
 
  printf("\n   Writing: %08x - %08x", &EXT_DRAM, EXT_DRAM_START + (EXT_DRAM_SIZE-1)); 
  
  // write then read entire array
  mem_ptr = &EXT_DRAM;
  for (x = 0; x < EXT_DRAM_WORDS-1; x+=2){
	  *mem_ptr = 0x55555555;
	  ++mem_ptr;
	  *mem_ptr = 0xAAAAAAAA;
	  ++mem_ptr;
  }

  printf("\n   Reading: %08x - %08x", &EXT_DRAM, EXT_DRAM_START + (EXT_DRAM_SIZE-1)); 
  mem_ptr = &EXT_DRAM;
  for (x = 0; x < EXT_DRAM_WORDS-1; x+=2){
	  rval = *mem_ptr;
	  if (rval != 0x55555555){
	    ++error;
	    printf("\naddr: %08x    wrote %08x  read %08x", mem_ptr, 0x55555555, rval); 
	  }
  	++mem_ptr;

	  rval = *mem_ptr;
	  if (rval != 0xAAAAAAAA){
	    ++error;
	    printf("\naddr: %08x    wrote %08x  read %08x", mem_ptr, 0xAAAAAAAA, rval); 
	  }
  	++mem_ptr;

  	if(error > 8){
	    printf("\nAborting test.");
  	  break;  // abort if too many errors
  	}
  }
  
  
  
  printf("\nTesting all 00000000.."); 
  mem_ptr = &EXT_DRAM;
  for (x = 0; x < EXT_DRAM_WORDS-1; x+=2){
	  *mem_ptr = 0x00000000;
	  rval = *mem_ptr;
	  if (rval != 0x00000000){
	    ++error;
	    printf("\naddr: %08x    wrote %08x  read %08x", mem_ptr, 0x00000000, rval); 
	  }
  	++mem_ptr;
  }

  printf("\nTesting all FFFFFFFF.."); 
  mem_ptr = &EXT_DRAM;
  for (x = 0; x < EXT_DRAM_WORDS-1; x+=2){
	  *mem_ptr = 0xFFFFFFFF;
	  rval = *mem_ptr;
	  if (rval != 0xFFFFFFFF){
	    ++error;
	    printf("\naddr: %08x    wrote %08x  read %08x", mem_ptr, 0xFFFFFFFF, rval); 
	  }
  	++mem_ptr;
  }
  
  printf("\nTesting 16-bit.."); 
  mem_ptr16 = &EXT_DRAM16;
  *mem_ptr16 = 0x8420;
  if(*mem_ptr16 != 0x8420){
    printf("\naddr: %08x    wrote %04x  read %04x", mem_ptr16, 0x8420, *mem_ptr16); 
  }
  
  
/*
  mem_ptr = &EXT_DRAM;

  printf("\n --- DDR CONFIG = %08x", MDDRC.DDR_SYS_CONFIG.R); 

  *mem_ptr = 0x00000000;
  printf("\naddr: %08x    wrote %08x  read %08x", mem_ptr, 0x00000000, *mem_ptr); 
	mem_ptr++;

  printf("\n --- DDR CONFIG = %08x", MDDRC.DDR_SYS_CONFIG.R); 

  *mem_ptr = 0xFFFFFFFF;
  printf("\naddr: %08x    wrote %08x  read %08x", mem_ptr, 0xFFFFFFFF, *mem_ptr); 
	mem_ptr++;

  printf("\n --- DDR CONFIG = %08x", MDDRC.DDR_SYS_CONFIG.R); 

  mem_ptr = &EXT_DRAM;
  *mem_ptr++ = 0x12345678;
  *mem_ptr =   0x9ABCDEF0;

  printf("\n --- DDR CONFIG = %08x", MDDRC.DDR_SYS_CONFIG.R); 

  mem_ptr = &EXT_DRAM;
  printf("\naddr: %08x    wrote %08x  read %08x", mem_ptr, 0x12345678, *mem_ptr++); 
  printf("\naddr: %08x    wrote %08x  read %08x", mem_ptr, 0x9ABCDEF0, *mem_ptr); 

  mem_ptr16 = &EXT_DRAM16;
  *mem_ptr16 = 0x8420;
  printf("\naddr: %08x    wrote %04x  read %04x", mem_ptr16, 0x8420, *mem_ptr16++); 
  *mem_ptr16 = 0xFFFF;
  printf("\naddr: %08x    wrote %04x  read %04x", mem_ptr16, 0xFFFF, *mem_ptr16++); 
  *mem_ptr16 = 0x0000;
  printf("\naddr: %08x    wrote %04x  read %04x", mem_ptr16, 0x0000, *mem_ptr16++); 
  *mem_ptr16 = 0x1234;
  printf("\naddr: %08x    wrote %04x  read %04x", mem_ptr16, 0x1234, *mem_ptr16++); 

  printf("\n --- DDR CONFIG = %08x", MDDRC.DDR_SYS_CONFIG.R); 
*/  
 
	if(error)  printf("\n\n--- DDR Memory Test:  %d  Errors\n", error);
	else			 printf("\nPASSED\n");
}


// --------------------------------------------------------------------------
// CLKOUT() - Outputs chosen sysclk
void CLKOUT(void){
    /* Configure CLKOUT on PB6 */
    SIU.PCR[22].R      = 0x0600;      		// PB3 clkout (0x0A00)

    /* Choose Clock source to be output */ 
//    CGM.OCDS_SC.B.SELCTL = 0x0; 	   	 	// 16MHz IRC Osc is Output Clock (Internal)
//    CGM.OCDS_SC.B.SELCTL = 0x1; 	   	 	// 4MHz crystal Osc is Output Clock (External crystal 40Mhz on my EVB)
    CGM.OCDS_SC.B.SELCTL = 0x2; 	   	 	// PLL0 is Output Clock (180MHz)
//    CGM.OCDS_SC.B.SELCTL = 0x3; 	   	 	// PLL1 is Output Clock (180MHz)

    /* Choose what to divide output by */
//    CGM.OCDS_SC.B.SELDIV = 0x0; 	   	 	// Divide Output Clock by 1 (No Division)
//    CGM.OCDS_SC.B.SELDIV = 0x1; 	   	 	// Divide Output Clock by 2
#ifdef CLOCK_SPEED_90MHZ
    CGM.OCDS_SC.B.SELDIV = 0x2; 	   	 	// Divide Output Clock by 4
#endif
#ifdef CLOCK_SPEED_180MHZ
    CGM.OCDS_SC.B.SELDIV = 0x3; 	   	 	// Divide Output Clock by 8
#endif

    /* Enable clock output */
    CGM.OC_EN.B.EN       = 0x1; 	   	  	// Enable Output Clock
}


// --------------------------------------------------------------------------
// setup_fmpll()
void setup_fmpll(void){

// Enable Clockout at Div 8 
 SIU.PCR[22].R      = 0x0600;        // Configure CLKOUT on B3 	

// Select 1 from below          
//    CGM.OCDSSC.B.SELCTL = 3; 	   	 	// PLL1 is Output Clock
    CGM.OCDS_SC.B.SELCTL = 2; 	   	 	// PLL0 is Output Clock 
//    CGM.OCDSSC.B.SELCTL = 1; 	   	 	// OSC is Output Clock
//    CGM.OCDS_SC.B.SELCTL = 0; 	   	 	// IRC is Output Clock 
    CGM.OCDS_SC.B.SELDIV = 3; 	   	 	// Divide Output Clock by 8
    CGM.OC_EN.B.EN       = 1; 	   	  	// Enable Output Clock

 ME.MER.R = 0x000025FF;          // Enable all modes                     		
 ME.RUNPC[0].R = 0x000000FE;     // Enable all peripherals in all modes  
 
 ME.DRUN.B.XOSCON = 1;			   // Switch on external osc 

// CGM.OCDS_SC.B.SELDIV = 3;
    
 // RE  enter the drun mode, to update the configuration 
 ME.MCTL.R = 0x30005AF0;                // Mode & Key 
 ME.MCTL.R = 0x3000A50F;                // Mode & Key inverted 

 while(ME.GS.B.S_MTRANS == 1){;}          // Wait for mode entry to complete 
 while(ME.GS.B.S_CURRENTMODE != 3){;}     // Check DRUN mode has been entered 

 while(!ME.GS.B.S_XOSC){;}		 // Wait for clock to stabilise 

 // Selecting XOSC as the FMPLL_0 Source 
 CGM.AC3_SC.B.SELCTL = 0x1;
  
 // For XTAL = 40MHz, VCO = 256M, PLL OUT = 16M 
 CGM.FMPLL[0].CR.B.IDF = 0x7;		// Divide by 8 
 CGM.FMPLL[0].CR.B.ODF = 0x0;		// Divide by 2 
 CGM.FMPLL[0].CR.B.NDIV = CLOCK_NDIV;		

 ME.DRUN.B.PLL0ON  = 1;

// RE enter the drun mode, to update the configuration 
 ME.MCTL.R = 0x30005AF0;     	        // Mode & Key 
 ME.MCTL.R = 0x3000A50F;                // Mode & Key inverted 

 while(ME.GS.B.S_MTRANS == 1){;}		// Wait for mode entry to complete 
 while(ME.GS.B.S_CURRENTMODE != 0x3){;}   // Check DRUN mode has been entered 

// wait for PLL to lock - will not lock before DRUN re-entry 
 while(CGM.FMPLL[0].CR.B.S_LOCK == 0){;}

 CGM.OCDS_SC.B.SELCTL = 2;		 // Select Output clock System PLL_0 

// Set the sysclock as PLL0 
 ME.DRUN.B.SYSCLK = 4; 
 
// RE enter the drun mode, to update the configuration 
 ME.MCTL.R = 0x30005AF0;     	        // Mode & Key 
 ME.MCTL.R = 0x3000A50F;                // Mode & Key inverted 

 while(ME.GS.B.S_MTRANS == 1){;}				// Wait for mode entry to complete 
 while(ME.GS.B.S_CURRENTMODE != 0x3){;}   // Check DRUN mode has been entered 
 
} 
/*
// --------------------------------------------------------------------------
// MC_MODE_INIT() - Initialize the Magic Carpet module
void MC_MODE_INIT(void)
{
   // Enable all peripheral clocks 
   CGM.SC_DC0.R = 0x80000000;  // Enable clock divider 
   // Setting RUN Configuration Register ME_RUN_PC[0] 
   ME.RUNPC[0].R=0x000000FE; // Peripheral ON in every mode 
   
   // Re-enter in DRUN mode to update 
   ME.MCTL.R = (DRUN_MODE << 28 | 0x00005AF0);        // Mode & Key 
   ME.MCTL.R = (DRUN_MODE << 28 | 0x0000A50F);        // Mode & Key 
   // Wait for mode entry to complete 
   while(ME.GS.B.S_MTRANS==1){;}      
   // Check DRUN mode has been entered 
   while(ME.GS.B.S_CURRENTMODE!=DRUN_MODE){;}
}


// --------------------------------------------------------------------------
void SET_PLL_0(void)
{
	CGM.CMU_0_CSR.B.RCDIV = 3; // Avoid CMU Reset if Fmosc < Firc
	CGM.OSC_CTL.B.EOCV = 0x20; // set the End of Count Value - time before the osc is made available to system. May not need for canned osc bypass mode 

	// --> Setup OSC
	CGM.AC3_SC.B.SELCTL =0x1;
	CGM.AC4_SC.B.SELCTL =0x1;
	ME.DRUN.R = 0x001F0030;    // Output Clock Division Select Register($C3FD_C02C)???
	
	// RE enter the drun mode, to update the configuration
	ME.MCTL.R = 0x30005AF0;    // Mode & Key
	ME.MCTL.R = 0x3000A50F;    // Mode & Key inverted
	while(ME.GS.B.S_XOSC == 0){};   // Wait for external osc to stabilize
	while(ME.GS.B.S_MTRANS == 1){};   // Wait for mode entry to complete
	while(ME.GS.B.S_CURRENTMODE != 3){}; // Check DRUN mode has been entered
	
	// --> Setup PLL
	// Max bus 180Mhz, VCO 256-512Mhz
	// Fpll = XTAL * NDIV / IDF / ODF = 180MHz
	// Fvco = XTAL * NDIV / IDF = 360MHz
	CGM.FMPLL[0].CR.B.IDF=0x7; 	// Divide by 8 for 40MHz XTAL 
	CGM.FMPLL[0].CR.B.ODF=0x0;  	// Divide by 2 
//	CGM.FMPLL[0].CR.B.NDIV=0x48; 	// Loop divide by 72 180Mhz 
 CGM.FMPLL[0].CR.B.NDIV = CLOCK_NDIV;		

	ME.DRUN.B.PLL0ON = 1;    // enable pll pll0

	// RE enter the drun mode, to update the configuration
	ME.MCTL.R = 0x30005AF0;    // Mode & Key
	ME.MCTL.R = 0x3000A50F;    // Mode & Key inverted
	while(ME.GS.B.S_MTRANS == 1){};   // Wait for mode entry to complete
	while(ME.GS.B.S_CURRENTMODE != 0x3){}; // Check DRUN mode has been entered

	while(CGM.FMPLL[0].CR.B.S_LOCK != 1){}; // Wait for PLL lock

	// --> Setup system clock
	ME.DRUN.B.SYSCLK = 0x4;    // system clock is sourced from PLL

	// RE enter the drun mode, to update the configuration
	ME.MCTL.R = 0x30005AF0;    // Mode & Key
	ME.MCTL.R = 0x3000A50F;    // Mode & Key inverted
	while(ME.GS.B.S_MTRANS == 1){};   // Wait for mode entry to complete
	while(ME.GS.B.S_CURRENTMODE != 0x3){}; // Check DRUN mode has been entered

	//ME_GS Poll Global status register to get current system clock
	//fail if system clock is not pll
	//0000 16MHz internal RC oscillator
	//0001 divided 16MHz internal RC oscillator
	//0010 4MHz crystal oscillator
	//0011 divided 4MHz crystal oscillator
	//0100 system PLL
	//1111 system clock is disabled
	while(ME.GS.B.S_SYSCLK != 4){}; // fail is stuck here  
}

// --------------------------------------------------------------------------
void SET_PLL_1()
{
	// --> Setup PLL
	// Max bus 180Mhz, VCO 256-512Mhz
	// Fpll = XTAL * NDIV / IDF / ODF = 180MHz
	// Fvco = XTAL * NDIV / IDF = 360MHz
	CGM.FMPLL[1].CR.B.IDF=0x7; 	// Divide by 8 for 40MHz XTAL 
	CGM.FMPLL[1].CR.B.ODF=0x0;  	// Divide by 1 
//	CGM.FMPLL[1].CR.B.NDIV=0x48; 	// Loop divide by 72 180Mhz 
 CGM.FMPLL[1].CR.B.NDIV = CLOCK_NDIV;		

	ME.DRUN.B.PLL1ON = 1;    // enable pll (mistake on header file)

	// RE enter the drun mode, to update the configuration
	ME.MCTL.R = 0x30005AF0;    // Mode & Key
	ME.MCTL.R = 0x3000A50F;    // Mode & Key inverted
	while(ME.GS.B.S_MTRANS == 1){};   // Wait for mode entry to complete
	while(ME.GS.B.S_CURRENTMODE != 0x3){}; // Check DRUN mode has been entered

	while(CGM.FMPLL[1].CR.B.S_LOCK != 1){}; // Wait for PLL lock
}
*/

// --------------------------------------------------------------------------
// Disable the software watchdog timer
void DISABLE_WATCHDOG(void){
  SWT_0.SR.R = 0x0000c520; // key 
  SWT_0.SR.R = 0x0000d928; // key 
  SWT_0.CR.R = 0x8000010A; // disable WEN 
}                                                                                                                                                                                                                                                                   


#define SLEW_HALF	0x0000
#define SLEW_FULL	0x0004

#define PCR_OUT	0x0200
#define PCR_IN	0x0100

#define PCR_A1	0x0400
#define PCR_A2	0x0800
#define PCR_A3	0x0C00

#define PCR_ODE	0x0020


//#define _SCR 0x0A8042C0 //dual	//RS=5; BS=4; 16BIT=0; RDLY=2; HALF=0; QURT=0; WDLY=2; EARLY=1; ODT=1
//#define _SCR 0x040042C0 
//#define _SCR 0x04004200 // removed OD
//#define _SCR 0x04005200 // extra half clock to read dly
#define _SCR 0x04004100 // extra clock to write dly
//#define _SCR 0x062042C0 // test #1
//#define _SCR 0x086042C0 // test #2
//#define _SCR 0x0AA042C0 // test #3

// Micron LPDDR SDRAM Commands
#define MIC_CS	0x00400000	// bit 22 CS#
#define MIC_RAS	0x00200000	// bit 21 RAS
#define MIC_CAS	0x00100000	// bit 20 CAS
#define MIC_WE	0x00080000	// bit 19 WE
#define MIC_BA2	0x00040000	// bit 18 BA2
#define MIC_BA1	0x00020000	// bit 17 BA1
#define MIC_BA0	0x00010000	// bit 16 BA0
// 14-0 = A14-A0

void Delay_NOP(unsigned long int val){
	while(val--){
		MDDRC.DDR_COMMAND.R = 0x00380000; // Issue NOP
	}
}

// --------------------------------------------------------------------------
void SetBusMode(char mode){

  vuint32_t *RegPtr;
  vuint16_t	val, i;

  val = 0x0100;		// make all bus ports INPUT, no pullups
  for(i=147;  i<=156; i++)  	SIU.PCR[i].R = val;
  for(i=158;  i<=197; i++)  	SIU.PCR[i].R = val;


	// --- DDR Mode
	if(mode){

		// disable EBI
	  RegPtr = (vuint32_t *) EBI_BASE_REG;  // MCR
  	*RegPtr = 0x00000010; // MDIS - module disabled
		
		// enable DDR: GPIO[129] = DDR ON	active high
    SIU.GPDO[128].B.PDO = 1;  	// make pins high, bus = 1.8V
    SIU.GPDO[129].B.PDO = 1;  

		// wait a microsecond before returning
		delay(100);	// ?? increase or decrease this?

	  // set PCR registers used by DDR
	  SIU.PCR[147].R = (PCR_A1 + PCR_OUT + SLEW_FULL);	// alt mode 1
	  SIU.PCR[148].R = PCR_IN;	// gpio input
	  SIU.PCR[149].R = PCR_IN;	// gpio input
	  for(i=150; i<=173; i++)  	SIU.PCR[i].R = (PCR_A1 + PCR_OUT + SLEW_FULL);	// alt mode 1
	  for(i=174; i<=193; i++)  	SIU.PCR[i].R = (PCR_A1 + PCR_OUT + PCR_IN + SLEW_FULL);	// alt mode 1


		MDDRC.DDR_SYS_CONFIG.R = 0xB0000000;	// Enable clock = no CKE

		// Configure DRAM Control Registers
		MDDRC.DDR_SYS_CONFIG.R = (0xB0000000 + _SCR);

		MDDRC.DDR_TIME_CONFIG0.R = 0x03C7261C;	//REFRESH=$3C7; CMD=$26; PRE=$1C
		MDDRC.DDR_TIME_CONFIG1.R = 0x40CC10E5;	//RFC=$10; WR1=$06; WTR1=$06; RRD=$02; RC=$07; RAS=$05
//org		MDDRC.DDR_TIME_CONFIG2.R = 0x23B10843;	//RCD=$02; FAW=$07; RTW1=$06; CCD=$02; RTP=$02; RP=$02; RPA=$03
		MDDRC.DDR_TIME_CONFIG2.R = 0x28321042;	

		delay(10000);	// wait c. 200us

		MDDRC.DDR_COMMAND.R = MIC_CS; // CS#
		delay(5);	

		// Set CKE high
		MDDRC.DDR_SYS_CONFIG.R = (0xF0000000 + _SCR);		// Enable CKE
		delay(20);	// Wait for 400ns
/*
		// NOP operation for 200us
		for(i=0; i<1000; i++){
			 MDDRC.DDR_COMMAND.R = 0x00380000; // Issue NOP
		}
*/
		Delay_NOP(1000);	// NOP operations for 200us


		// Send commands directly to DRAM

	// 0-14 = A0 - A14
	// 15 = 0
	// 16-18 = BA0 - BA2
	// 19 = WEB
	// 20 = CAS
	// 21 = RAS
	// 22 = CS0

		// precharge all command
		MDDRC.DDR_COMMAND.R = 0x00100400; // Issue PreCharge All
		Delay_NOP(5);	

		// refresh command 
		MDDRC.DDR_COMMAND.R = 0x00080000; // Issue Refresh
		Delay_NOP(5);	

		// refresh command
		MDDRC.DDR_COMMAND.R = 0x00080000; // Issue Refresh
		Delay_NOP(5);	

		// standard mode register command 
//		MDDRC.DDR_COMMAND.R = 0x00000033; 
		MDDRC.DDR_COMMAND.R = 0x00000032; 
		Delay_NOP(5);	

		// extended mode register command 
	MDDRC.DDR_COMMAND.R = (0x00000000 + MIC_BA1); 	// full drive
//	MDDRC.DDR_COMMAND.R = (0x00000020 + MIC_BA1); 	// half drive
		Delay_NOP(5);	


	
/* ---- original from greenhill project ----	

		// NOP operation for 200us
		for(i=0; i<1000; i++){
			 MDDRC.DDR_COMMAND.R = 0x00380000; // Issue NOP
		}
		
		// Set CKE high
		MDDRC.DDR_SYS_CONFIG.R = 0xF0000000 + _SCR;		// Enable CKE
		delay(20);	// Wait for 400ns

		// Send a precharge all command and thereafter wait at least tRP.
		MDDRC.DDR_COMMAND.R = 0x00100400; // Issue PreCharge All
		delay(5);

		// Send a load mode command to program EMR2 and thereafter wait at least tMRD.
		MDDRC.DDR_COMMAND.R = 0x00020080; // Issue LM to EMR2 - 1 x refresh rate
		delay(5);

		// Send a load mode command to program EMR3 and thereafter wait at least tMRD.
		MDDRC.DDR_COMMAND.R = 0x00030000; // Issue LM to EMR3 - no entries here
		delay(5);

		// Send a load mode command to EMR to enable DLL and thereafter wait at least tMRD.
		MDDRC.DDR_COMMAND.R = 0x00010002; // Issue LM to EMR to enable DLL  ; reduced drive
		delay(5);

		// Send a load mode command to program MR with reset DLL and thereafter wait at least tMRD.
		MDDRC.DDR_COMMAND.R = 0x00000332; // Issue LM to MR to reset DLL; Write recovery = 2; CL=3; Burst length=4
		delay(5);

		// Send a precharge all command and thereafter wait at least tRP.
		MDDRC.DDR_COMMAND.R = 0x00100400; // Issue PreCharge All
		delay(5);

		// Send a refresh command and thereafter wait at least tRP.
		MDDRC.DDR_COMMAND.R = 0x00080000; // Issue Refresh
		delay(5);

		// Send a refresh command and thereafter wait at least tRP
		MDDRC.DDR_COMMAND.R = 0x00080000; // Issue Refresh
		delay(5);

		// Send a load mode command to program the mode register without DLL reset (A8=0) and thereafter wait at least tMRD.
		MDDRC.DDR_COMMAND.R = 0x00000232; // Issue LM to MR to stop DLL  reset;; Write recovery = 2; CL=3 ; Burst length=4
		delay(5);

		// Wait until 200 clock cycles has passed after step 11 (DLL reset).
		delay(200);

		// Send a load mode command to EMR to enable OCD default and thereafter wait at least tMRD.3
		MDDRC.DDR_COMMAND.R = 0x00010382; // Issue LM to EMR to enable OCD default; reduced drive
		delay(5);

		// Send a load mode command to EMR to enable OCD exit and thereafter wait at least tMRD.3    
		if ((_SCR & (0x0000000C0)) > 0){
			//MDDRC.DDR_COMMAND.R = 0x00010046; // Issue LM to EMR to enable OCD exit; all other settings default; reduced drive - ODT=50ohm
	    MDDRC.DDR_COMMAND.R = 0x00010044; // Issue LM to EMR to enable OCD exit; all other settings default; full drive - ODT=50ohm
		}else{
	    MDDRC.DDR_COMMAND.R = 0x00010002; // Issue LM to EMR to enable OCD exit; all other settings default; reduced drive
		}
		delay(5);

*/
		MDDRC.DDR_SYS_CONFIG.R = (0xE0000000 + _SCR);	//End Command
		
		delay(100);	// ?? increase or decrease this?
	}

	// --- EBI Mode
	else{
	  // set PCR registers used by EBI 
	  for(i=147;  i<=154; i++)  	SIU.PCR[i].R = PCR_A2 + PCR_OUT + SLEW_FULL;	// alt mode 2, output
	  SIU.PCR[155].R = PCR_OUT;						// gpio output
	  SIU.PCR[156].R = PCR_A2 + PCR_OUT + SLEW_FULL;	// alt mode 2, output
    // no 157 
	  SIU.PCR[158].R = PCR_A3 + PCR_OUT;	// alt mode 3, output
	  SIU.PCR[159].R = PCR_IN;						// gpio input
//ex_mem_works	  SIU.PCR[160].R = PCR_A3 + PCR_OUT + PCR_IN + PCR_ODE;	// alt mode 3, input & output, open drain
//ex_mem_works	  		SIU.PCR[161].R = PCR_A3 + PCR_OUT + PCR_IN + PCR_ODE;	// alt mode 3, input & output, open drain
//	  SIU.PCR[160].R = PCR_A3 + PCR_IN + PCR_ODE + 0x03;	// (TA) 
	  SIU.PCR[160].R = PCR_A3 + PCR_IN + 0x03;	// (TA) 
	  SIU.PCR[161].R = PCR_A3 + PCR_IN + 0x03;	// (TEA)

	  for(i=162;  i<=165; i++)  	SIU.PCR[i].R = PCR_IN;	// gpio input
	  for(i=166;  i<=197; i++)  	SIU.PCR[i].R = PCR_A2 + PCR_OUT + PCR_IN + SLEW_FULL;	// alt mode 2, input & output

		// GPIO[128] = EBI ON	active low
    SIU.GPDO[128].B.PDO = 0;  	// make pins low, bus = 3V
    SIU.GPDO[129].B.PDO = 0;  

		// wait a microsecond before returning
		delay(100);	// ?? increase or decrease this?

		// setup EBI registers
  	RegPtr = (vuint32_t *) EBI_BASE_REG;  // MCR
  	*RegPtr = 0x00000006; // NON CS access = 32-bit muxed mode, D16-D31 = 16bit 
	}
}

// --------------------------------------------------------------------------
// Setup EBI

void setup_EBI(){
  vuint32_t *RegPtr;
 

  SIU.GPDO[128].B.PDO = 0;  	// make pins low, bus = 3V
  SIU.GPDO[129].B.PDO = 0;  

	// make bus mode pins GPIO outputs
	SIU.PCR[128].R = 0x0200;	
	SIU.PCR[129].R = 0x0200;	

	SetBusMode(EBI_BUSMODE);	// Set bus to EBI mode

//  RegPtr = (vuint32_t *) (EBI_BASE_REG + 0x0C);  // BMCR
//  *RegPtr = 0x00;  // disable bus monitor
 
    
  // config the base register: Address 2000 0000
  RegPtr = (vuint32_t *) (EBI_BASE_REG + 0x14);  	// bank 0
  *RegPtr = 0xFF800010; // OR0    1 wait
//  *RegPtr = 0xFF800000; // OR0      0 wait
  RegPtr = (vuint32_t *) (EBI_BASE_REG + 0x10);  	// bank 0
//  *RegPtr = 0x200000AF; // BR0    //--- CONFIGURE CS0 - 20000000 - 207FFFFF
//  *RegPtr = 0x200000AB; // BR0    //--- CONFIGURE CS0 - 20000000 - 207FFFFF
  *RegPtr = 0x200010AB; // BR0    //--- CONFIGURE CS0 - 20000000 - 207FFFFF

  RegPtr = (vuint32_t *) (EBI_BASE_REG + 0x24);  	// bank 2
  *RegPtr++ = 0xFF8000F0; // OR2    
  RegPtr = (vuint32_t *) (EBI_BASE_REG + 0x20);  	// bank 2
  *RegPtr++ = 0x208008C3; // BR2    //--- CONFIGURE CS2 - 20800000 - 20FFFFFF


//  *RegPtr++ = 0x208000E3; // BR1  // unused
//  *RegPtr++ = 0xFF800000; // 0R1    
//  *RegPtr++ = 0x218000E3; // BR3  // unused
//  *RegPtr++ = 0xFF800000; // 0R3    

  *RegPtr-- = 0xFF8000F0;   // OR1
  *RegPtr = 0x208008C3; // BR1    


  // set CLKOUT divide by 4 (144MHz / 4 = 36MHz 
  //SIU.ECCR.R |= 0x3;


/*

  // from mpc5633m_booke.mac
  //REM Set up MMU for External Memory
  //REM Base address = $2000_0000
  //REM TLB2, 16 MByte Memory Space, Not Guarded, Cachable, All Access
  //spr 624t $1002 0000 ; MAS0
  //spr 625t $C000 0700 ; MAS1
  //spr 626t $2000 0000 ; MAS2
  //spr 627t $2000 003F ; MAS3
  //execute_opcode $7C0007A4 ; tlbwe
  asm{
     // load r3 with MAS0 value
     lis r3,0x1002
     
     // load r4 with MAS1 value
     lis r4,0xC000
//     ori r4,r4,0x0780 //  --- 32MB
     ori r4,r4,0x0800 //  --- 64MB
     
     // load r5 with the external RAM address 
     lis r5,0x2000
     
     // end MAS3 setup in r4 
     lis r6,0x2000
     ori r6,r6,0x003F
     
     // write the MMU entry defined through r3, r4, r5 and r6  

    // Write MMU Assist Register 0 (MAS0); SPR 624 
    mtspr   624, r3
    // Write MMU Assist Register 1 (MAS1); SPR 625 
    mtspr   625, r4
    // Write MMU Assist Register 2 (MAS2); SPR 626 
    mtspr   626, r5
    // Write MMU Assist Register 3 (MAS3); SPR 627 
    mtspr   627, r6

    // Write the table entry 
    tlbwe
  }

  //-------------------------------
  // setup pads 

	// default to primary function, pull ups
  for (x = 256; x <= 307; x++){
//	  SIU.PCR[x].R = 0x04C3;  // max pad drive, 
	  SIU.PCR[x].R = 0x0403;  
  }

  // rewrite these PCR as they are DAT function in 32-bit muxed mode 
  for (x = 263; x <= 277; x++){
//	  SIU.PCR[x].R = 0x08C3;  // function 2, max pad drive, pull ups
	  SIU.PCR[x].R = 0x0803;  // function 2, pull ups
  }

  // setup other EBI pins 
  SIU.PCR[256].R = 0x0400;    // CS0 
  SIU.PCR[257].R = 0x0803;    // D31 
  SIU.PCR[258].R = 0x0400;    // CS3 
  SIU.PCR[294].R = 0x0400;    // RD_~WR 
  SIU.PCR[295].R = 0x0400;    // ~WE0 
  SIU.PCR[296].R = 0x0400;    // ~WE1 
  SIU.PCR[297].R = 0x0400;    // ~OE  
  SIU.PCR[298].R = 0x0400;    // ~TS 
  SIU.PCR[302].R = 0x0400;    // ~BDIP 
  SIU.PCR[303].R = 0x0400;    // ~WE2  
  SIU.PCR[304].R = 0x0400;    // ~WE3  

*/
  
}


void initialize_hardware(){

  vuint16_t	val, i;
  
  val = 0x0103;	// make INPUT, push pull by default

	// set default GPIO for all pins
  for(i=0;  i<=34; i++)  	SIU.PCR[i].R = val;
  for(i=36; i<=39; i++)  	SIU.PCR[i].R = val;
  for(i=42; i<=60; i++)  	SIU.PCR[i].R = val;
  SIU.PCR[62].R = val;
  SIU.PCR[64].R = val;
  SIU.PCR[66].R = val;
  for(i=68; i<=71; i++)  	SIU.PCR[i].R = val;
  for(i=73; i<=76; i++)  	SIU.PCR[i].R = val;
  SIU.PCR[78].R = val;
  SIU.PCR[80].R = val;
  for(i=84; i<=95; i++)  	SIU.PCR[i].R = val;
  for(i=98; i<=103; i++) 	SIU.PCR[i].R = val;
  for(i=108; i<=146; i++)	SIU.PCR[i].R = val;
	// between these used by data bus
  for(i=198; i<=233; i++)	SIU.PCR[i].R = val;


//	MC_MODE_INIT();		// set magic carpet 

	setup_fmpll();	// Setup PLL 
//	SET_PLL_0();		// Set PLL 0
//	SET_PLL_1();		// Set PLL 1
	CLKOUT();		// Output Selected Clock Source 
	
//	SIU.PCR[148].R = 0x0220;  //FLEXPWM0_B0 Ball=K17
	
	sci_a_init();	// initialize SCI_A serial port (1st core uses SCI_A)

  setup_EBI();	// initialize the external bus interface and chip selects
  
}